#----------------------------------------------------------------------------#
# Copyright 2010, Regular Polygon

# Permission to use this software without fee is hereby granted.
# Distribution or publication of this software, or substantial portions of
# the software, is subject to the expressed written consent of the author.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#----------------------------------------------------------------------------#
# Name        :   Sierpinski Tetrahedron
# Filename    :   rp_sierpinski_tetrahedron.rb
# Version     :   1.0
# SU Version  :   6.0+
# Type        :   Dialog Box
# Description :   Draws a Sierpinski Tetrahedron, levels 0-7. 
# Menu Item   :   Draw -> Sierpinski Tetrahedron
# Context Menu:   N/A
# Usage       :   Fill in the dialog box.
# Date        :   8/16/2010
# Author      :   Regular Polygon
# Email       :   regular.polygon@gmail.com
# Website     :   http://regularpolygon.blogspot.com/
#----------------------------------------------------------------------------#


require 'sketchup.rb'
module RP; end


#----------------------------------------------------------------------------#
module RP::SierpinskiTetrahedron

	hues = ["red", "DarkOrange", "yellow", "Green", "Blue", "BlueViolet", "Magenta"]  
	@@color = hues.collect {|c|  Sketchup::Color.new(c).blend("white", 0.9) }

	# Default dialog box parameters
	@@height = (Sketchup.read_default 'RP_sierpinski_tetra', 'height', '42').to_f.inch
	@@level = (Sketchup.read_default 'RP_sierpinski_tetra', 'level', '3').to_i

	# points, and face to points association, for a unit tetrahedron
	@@tetra_pts = [[+1, +1, +1], [-1, -1, +1], [+1, -1, -1], [-1, +1, -1]]
	@@face_indices = [[0, 1, 2], [0, 3, 1], [0, 2, 3], [1, 3, 2]] 

	# @@tr rotates the tetrahedron to point up
	rz = Geom::Transformation.new(ORIGIN, Z_AXIS, 45.degrees)
	phi = Z_AXIS.angle_between [1, 1, 1]
	rx = Geom::Transformation.new(ORIGIN, X_AXIS, -phi)
	@@tr = rx * rz
	@@tr_inv = @@tr.inverse


	# Component definition for a unit tetrahedron
	def self.tetrahedron

		# Get the definition if it exists
		d = Sketchup.active_model.definitions['RP Tetrahedron 0']
		return d  if d   

		# Doesn't exist, create definition
		d = Sketchup.active_model.definitions.add 'RP Tetrahedron 0'
		entities = d.entities
		@@face_indices.each {|indices|  
			pts = indices.collect {|i| @@tetra_pts[i] } 
			entities.add_face(pts)
		}

		d
	end


	# 4 transformations that build level n from level n-1 sub-components
	@@trans = [] 
	scale = Geom::Transformation.new(0.5)
	@@tetra_pts.each {|pt|
		translate = Geom::Transformation.new(pt)
		@@trans.push(scale * translate) 
	}


	# Component definition for level n of Sierpinski Tetrahedron
	def self.component(n)

		# Base case level 0
		return self.tetrahedron  if ( n == 0 )

		# Get the definition if it exists
		d = Sketchup.active_model.definitions["RP Sierpinski Tetrahedron #{ n }"]
		return d  if d   

		# Doesn't exist, create the definition
		d = Sketchup.active_model.definitions.add "RP Sierpinski Tetrahedron #{ n }"
		entities = d.entities

		# build the definition from subcomponents
		d1 = self.component(n-1) 
		components = []
		@@trans.each {|t| components.push( entities.add_instance(d1, t) ) }

		# explode them and regroup geometry by material
		components.each {|comp| comp.explode }

		# Collect inside faces
		faces = get_inside_faces(entities)
		# and copy them to a component.
		d1 = self.copy_faces(faces, "RP Sierpinski Tetrahedron #{ n }-#{ n }")

		# delete the original faces and edges
		entities.erase_entities faces
		edges = entities.find_all {|e|  e.kind_of?(Sketchup::Edge) && e.faces.length == 0 }
		entities.erase_entities edges

		# add a sub-component with same geometry and color it
		inst = entities.add_instance(d1, @@tr_inv)
		inst.material = @@color[n-1] 

		# Now group the level n-1 subcomponents leftover from the explosion.  
		m = n - 1
		while (m > 0)
			# collect the instances of the m-th subcomponent
			d1 = Sketchup.active_model.definitions["RP Sierpinski Tetrahedron #{ n-1 }-#{ m }"]
			components = d1.instances.find_all {|inst|  inst.parent == d }

			# Create a new component definition, and copy each instance to it.
			d2 = self.copy_components(components, "RP Sierpinski Tetrahedron #{ n }-#{ m }")

			# delete the original components
			entities.erase_entities components

			# add a sub-component with the copies, and make it the same color
			inst = entities.add_instance(d2, @@tr_inv)
			inst.material = @@color[m-1]  

			m -= 1
		end

		d
	end


	# collect inside faces.
	def self.get_inside_faces(entities)
		faces = []

		entities.each {|f|
			next  unless ( f.kind_of?(Sketchup::Face) )
			inside = true
			f.vertices.each {|v| 
				pt = v.position
				sum = pt.x.abs + pt.y.abs + pt.z.abs
				if ( sum != 1 )
					inside = false 
					break
				end
			}
			faces.push f  if inside
		}

		faces
	end


	# Create a component definition with a copy of the faces
	def self.copy_faces(faces, def_name)

		# Get the definition if it exists
		d1 = Sketchup.active_model.definitions[def_name]
		return d1  if d1   

		# Doesn't exist, create the definition
		d1 = Sketchup.active_model.definitions.add def_name

		# copy faces
		for f in faces
			pts = f.outer_loop.vertices.collect {|v| @@tr * v.position }
			d1.entities.add_face pts
		end

		d1
	end


	# Create a component definition with a copy of component instances.
	def self.copy_components(components, def_name)

		# Get the definition if it exists
		d1 = Sketchup.active_model.definitions[def_name]
		return d1  if d1   

		# Doesn't exist, create the definition, and copy the instances
		d1 = Sketchup.active_model.definitions.add def_name
		components.each {|inst|  
			d1.entities.add_instance(inst.definition, @@tr * inst.transformation) 
		}

		d1
	end


	# This method handles the Draw -> Sierpinski Tetrahedron menu item.
	def self.dialog

		# Show dialog to get user input for Sierpinski Tetrahedron paramters
		prompts = ['Height: ', 'Level: ']
		defaults = [@@height, @@level]
		list = [nil, '0|1|2|3|4|5|6|7']
		results = nil

		# Keep prompting user until input is valid.
		begin
			results = UI.inputbox prompts, defaults, list, 'Sierpinski Tetrahedron'
			return unless results  # user cancelled the operation
			height, level = results
			raise "Invalid height: #{ height }"  if ( height <= 0 )
		rescue
			UI.messagebox $!.message
			retry
		end
		
		# Remember parameters
		@@height, @@level = results

		# Remember parameters between sessions
		Sketchup.write_default 'RP_sierpinski_tetra', 'height', @@height.inspect
		Sketchup.write_default 'RP_sierpinski_tetra', 'level', @@level.to_s

		self.draw(@@level)
	end

	
	def self.draw(n)

		# make component definition creation undoable
		Sketchup.active_model.start_operation 'Sierpinski Tetrahedron' 

			d = Sketchup.active_model.definitions.add 'Sierpinski Tetrahedron'
			entities = d.entities

			d1 = self.component(n)
			s = @@height * Math.sqrt(3) / 4
			scale = Geom::Transformation.new(s)
			t = @@tr * scale 
			inst = entities.add_instance(d1, t)
			inst.explode

			# Move insertion point to a bottom corner
			x = -Math.sqrt(2) * s
			y = x / Math.sqrt(3)
			z = -s / Math.sqrt(3)
			d.insertion_point = Geom::Point3d.new(x, y, z)

		Sketchup.active_model.commit_operation 

		# place instance in model
		Sketchup.active_model.place_component( d )
	end

end  # module RP::SierpinskiTetrahedron


# Menu
#----------------------------------------------------------------------------#
unless file_loaded? File.basename(__FILE__) 
	# add_separator_to_menu("Draw")
	# UI.menu("Draw").add_item("Sierpinski Tetrahedron") { RP::SierpinskiTetrahedron.dialog } 
	path = Sketchup.find_support_file("Plugins/rp_sierpinski_tetrahedron")
  SierpinskiTetrahedron_tb = UI::Toolbar.new("谢尔宾斯基三角形海绵体")
  cmd1 = UI::Command.new("谢尔宾斯基三角形海绵体") { RP::SierpinskiTetrahedron.dialog } 
  cmd1.small_icon = File.join(path,"icons/SierpinskiTetrahedron24.jpg")
  cmd1.large_icon = File.join(path,"icons/SierpinskiTetrahedron36.jpg")
  cmd1.tooltip = "谢尔宾斯基三角形海绵体"
  cmd1.status_bar_text = "谢尔宾斯基三角形海绵体"
  SierpinskiTetrahedron_tb.add_item(cmd1)
  SierpinskiTetrahedron_tb.show
end

file_loaded File.basename(__FILE__)